// =========================
// share_app_ttcb.h
// =========================

#ifndef SHARE_TTCB
#define SHARE_TTCB
#include "../../crypto/rsa/mpi.h"
#include "../../crypto/rsa/rsa.h"
#include "share_app_tcb.h"


/*
+------------------------------------+
|   The values for LOG / DEBUG       |
|                                    |
|  Change these if you want verbose  |
+------------------------------------+
*/

/*
int LOG_LEVEL = 0;
int DBG_LEVEL = 0;
*/

/*
+--------------------------------+
| TTCB Public Key                |
+--------------------------------+
*/

//RSA_public_key TTCBPublicKey;


/*
+--------------------------------+
| ERRORS RETURNED                |
+--------------------------------+
*/

#define PARAM_OUT_OF_BOUND	-41	// parameter out of bound 
#define NO_ROOM_FOR_ENTITY	-42	// no room for new entities in this local TTCB
#define EID_NOT_IN_ELIST	-43	// (agreement)
#define TSTART_EXPIRED		-44	// (agreement)
#define MAX_NUM_AGREEMENTS_RUNNING	-45	// (agreement)
#define VALUE_ALREADY_PROPOSED -46	// an entity is trying to propose a value to a agreement to which it already did
#define INVALID_TAG		-47
#define AGREEMENT_RUNNING 	-48	// (agreement)
#define INVALID_ID		-49	// invalid eid, wrong secret, ...
#define INVALID_SEQ		-50	// invalid sequence number
#define INVALID_DECISION	-51	// (agreement)
#define DID_NOT_PROPOSE		-52	// (agreement)
#define NO_RESOURCES_PROPOSE	-53	// (agreement -- no resources for a propose)
#define INVALID_PROTECTION		-54	// invalid protection parameter
#define FAILED_LOCAL_AUTHENTICATION     -55     // Local authentication failed
#define AGREEMENT_EXPIRED	-56	// (agreement) the agreement results were discarded

// Some constants
#define	MAX_LOCAL_ENTITIES 20 // maximum n. of entities in a local TTCB
#define	MAX_SIZE_ELIST	10  // changed => may require to change mask_t size!
#define	MAX_AGREEMENTS	100  // max num of simultaneous agreements
#define	MAX_FIELDS_CSTABLE 10
#define TIME_DISCARD_AGREEMENT_RESULTS 1000000000  // in nanoseconds (1 second for now)


// protection types
#define	PROTECTION_NONE	0
#define	PROTECTION_HMD5	1

#define END_ELIST 0xFFFF   // 0xFFFF



// Agreement decisions
#define	TTCB_TBA_RMULTICAST	0
#define	TTCB_TBA_MAJORITY	1
#define	TTCB_TBA_AND		2
#define TTCB_TBA_OR             3
#define TTCB_TBA_XOR            4

#define	MAX_NUMBER_DECISIONS	4	// The maximum number of existing decisions


// --------------
// MACROS
// --------------

// eid (16 bits) =  | ltid  | index |      (local TTCB id + local eid)
#define EIDtoLTID(e)   ((e) / 256)	// converts an eid to a local TTCB id
#define EIDtoINDEX(e)  ((e) & 0x00FF)	// converts an eid to the index
#define toEID(l,i)     ((l) * 256 + i)	// converts an index and a local TTCB id to an eid


/* Sizes of the types (in bytes): char-1 short-2 int-4 long-4  long long-8  long double-12 */
     

// -------------------
// TYPES
// -------------------

typedef unsigned char bits64[8];
typedef unsigned char bits128[16];
typedef unsigned char bits160[20];
typedef unsigned char bits512[64];
typedef unsigned char bits1024[128];

typedef unsigned short eid_t;	// eid type (2 bytes)
typedef bits128 skey_t;		// symmetric key type
typedef bits128 akey_t;		// asymmetric key type
typedef unsigned char sign_t[152]; 	// local TTCB private key signature
typedef char protection_t;		// protection algorithm
typedef bits64 challenge_t;	// challenge (LA)
typedef char decision_t;	// decision type (Agreement)
typedef bits160 value_t; 	// value type (Agreement)
typedef short tag_t;		// tag type (2 bytes)
typedef unsigned char localTTCBid_t;
typedef unsigned char seq_t;	// sequence number type
typedef bits128 random_t;	// random number type
typedef eid_t elist_t[MAX_SIZE_ELIST];	// elist type (last record has value END_ELIST)
typedef short mask_t;		// 1 bit per elist member (type depends on MAX_SIZE_ELIST)



// descriptor used in every TTCB call
// (required to implement the secure channel; the programmer 
//   should use this transparently)
typedef struct {
	int desc_orig, wi_orig; // desc and wi as in localAuthentication
	eid_t eid;
	protection_t protection;
	skey_t secret;
	seq_t seq;
	challenge_t challenge;
	RSA_public_key TTCB_PublicKey;
} descriptor_t;



// ------------------------------------------------
// TTCB functions in and out parameter structures
// ------------------------------------------------

// HEADERS 

// all *in* structures MUST have the following in the beggining (except local authent)
typedef struct {
	eid_t eid;	// never encrypted
	eid_t unused;	// for 32 bit alignment (also not encrypted)
	skey_t secret_mic;	// shared secret if authenticity; MIC if integrity but not confidentiality(also not encrypted)
	eid_t copy_eid;	// encrypted if confidentiality or integrity requested
	seq_t seq;	// sequence number
} in_hdr_t;

// NOTE: the way in which the secure channel operations are performed requires the first 4 variables of in_hdr_t to have this order and this sizes

// all *out* structures MUST have the following in the beggining (except local authent)
typedef struct {
	skey_t secret_mic; // never encrypted
	eid_t eid;
	seq_t seq;
	char error;
} out_hdr_t;

// NOTE: the way in which the secure channel operations are performed requires the first 2 variables of out_hdr_t to have this order and this sizes


#define MAX_SIZE_OUT_STRUCT 128  // this must be larger than any out struct

// STRUCTURES

typedef struct {
	skey_t key;
	protection_t protection;
	challenge_t challenge;
	char tail[39];	// important! see calculate_tails.c 
} LA_in_t;

typedef struct {
	sign_t challenge_sign;
	eid_t eid;
	char error;
	char tail[37];
} LA_out_t;

typedef struct {
	eid_t eid, unused;
	skey_t secret_mic;
	eid_t copy_eid;
	seq_t seq;
	elist_t elist;
	TCBtimestamp tstart;
	decision_t decision;
	value_t value;
	char tail[8];
} propose_in_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	tag_t tag;
	char tail[58];
} propose_out_t;

typedef struct {
        eid_t eid, unused;
        skey_t secret_mic;
        eid_t copy_eid;
        seq_t seq;
	tag_t tag;
	char tail[58];
} decide_in_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	value_t value;
	mask_t mask1, mask2;
	char tail[36];
} decide_out_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	random_t random;
	char tail[44];
} rand_out_t;

typedef struct {
        eid_t eid, unused;
        skey_t secret_mic;
        eid_t copy_eid;
        seq_t seq;
	localTTCBid_t localTTCBid;
	char tail[60];
} localTTCBid_in_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	akey_t key;
	char tail[44];
} key_out_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	localTTCBid_t localTTCBid;
	char tail[58];
} localTTCBid_out_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	char tail[60];
} error_out_t;

typedef struct {
        eid_t eid, unused;
        skey_t secret_mic;
        eid_t copy_eid;
        seq_t seq;
	char tail[60];
} eid_in_t;

typedef struct {
	skey_t secret_mic;
	eid_t eid;
	seq_t seq;
	char error;
	TCBtimestamp ts;
	char tail[52];
} timestamp_out_t;

// ------------------------------------------------
// Security Wrappers for TCB services
// ------------------------------------------------


typedef struct{
	in_hdr_t	header_in;
	char buffer	[84-sizeof(in_hdr_t)];
}tcb_service_in;

typedef struct{
	in_hdr_t	header_in;
	char buffer	[168-sizeof(in_hdr_t)];
}tcb_service_big_in;

typedef struct{
	out_hdr_t	header_out;
	char buffer	[80-sizeof(out_hdr_t)];
}tcb_service_out;



#endif


